@@ -1,8 +1,7 @@
-
 /*
  * FCRON - periodic command scheduler 
  *
- *  Copyright 2000-2001 Thibault Godouet <fcron@free.fr>
+ *  Copyright 2000-2007 Thibault Godouet <fcron@free.fr>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,25 +21,35 @@
  *  `LICENSE' that comes with the fcron source distribution.
  */
 
- /* $Id: database.c,v 1.49 2001/02/27 20:50:22 thib Exp $ */
+ /* $Id: database.c,v 1.79 2007/04/14 18:04:18 thib Exp thib $ */
 
 #include "fcron.h"
 
+#include "database.h"
+#include "job.h"
+#include "getloadavg.h"
+
 int is_leap_year(int year);
 int get_nb_mdays(int year, int mon);
 void set_wday(struct tm *date);
-void goto_non_matching(CL *line, struct tm *tm, char option);
-void run_normal_job(CL *line);
+time_t mktime_no_dst(struct tm *t);
+void goto_non_matching(cl_t *line, struct tm *tm, char option);
+#define END_OF_INTERVAL 1    /* goto_non_matching() : option */
 void run_serial_job(void);
 void run_lavg_job(int i);
-void run_queue_job(CL *line);
+void run_queue_job(cl_t *line);
+void resize_exe_array(void);
 
 
+#ifndef HAVE_SETENV
+char env_tz[PATH_LEN];
+#endif
+
 void
 test_jobs(void)
   /* determine which jobs need to be run, and run them. */
 {
-    struct job *j;
+    struct job_t *j;
 
 /*      // */
     debug("Looking for jobs to execute ...");
@@ -48,7 +57,7 @@ test_jobs(void)
 
     while ( (j=queue_base) && j->j_line->cl_nextexe <= now ){
 	if ( j->j_line->cl_remain > 0 && --(j->j_line->cl_remain) > 0) {
-	    set_next_exe(j->j_line, STD);
+	    set_next_exe(j->j_line, STD, -1);
 	    debug("    cl_remain: %d", j->j_line->cl_remain);
 	    continue ;
 	}
@@ -56,32 +65,142 @@ test_jobs(void)
 	j->j_line->cl_remain = j->j_line->cl_runfreq;
 
 	if ( is_lavg(j->j_line->cl_option) )
-	    add_lavg_job(j->j_line);
+	    add_lavg_job(j->j_line, -1);
 	else if ( is_serial(j->j_line->cl_option) )
-	    add_serial_job(j->j_line);
+	    add_serial_job(j->j_line, -1);
 	else
-	    run_normal_job(j->j_line);
+	    run_normal_job(j->j_line, -1);
 	
-	set_next_exe(j->j_line, STD);
+	set_next_exe(j->j_line, STD, -1);
     }
 
 }
 
 
+int
+switch_timezone(const char *orig_tz, const char* dest_tz)
+/* check if we have already switched to dest_tz timezone, otherwise do it */
+/* If dest_tz is NULL, this function does nothing */
+/* Returns 1 if this function has switched the timezone, 0 otherwise */
+{
+    char *current_tz = getenv("TZ");
+	if ( dest_tz != NULL && 
+	     (current_tz == NULL || strcmp(dest_tz, current_tz) != 0)) {
+#ifdef HAVE_SETENV
+	    if ( setenv("TZ", dest_tz, 1) < 0 )
+		error_e("could not set env var TZ to %s", dest_tz);
+#else
+	    snprintf(env_tz, sizeof(env_tz) - 1, "TZ=%s", dest_tz);
+	    env_tz[sizeof(env_tz)-1] = '\0';
+	    if ( putenv(env_tz) < 0 )
+		error_e("could not set env var TZ to %s", dest_tz);
+#endif
+	    return 1;
+	}
+	else
+	    return 0;
+}
+
 void
-run_normal_job(CL *line)
+switch_back_timezone(const char *orig_tz)
+/* if orig_tz is NULL, unsets TZ
+ * otherwise, sets TZ to orig_tz */
+{
+    if ( orig_tz == NULL) {
+#ifdef HAVE_SETENV
+	unsetenv("TZ");
+#else
+	env_tz[0] = '\0';
+	if ( putenv(env_tz) < 0 )
+	    error_e("could not flush env var TZ");	
+#endif
+    }
+    else {
+#ifdef HAVE_SETENV
+	if ( setenv("TZ", orig_tz, 1) < 0 )
+	    error_e("could not set env var TZ to %s", orig_tz);
+#else
+	snprintf(env_tz, sizeof(env_tz) - 1, "TZ=%s", orig_tz);
+	env_tz[sizeof(env_tz)-1] = '\0';
+	if ( putenv(env_tz) < 0 )
+	    error_e("could not set env var TZ to %s", orig_tz);
+#endif
+    }
+}
+
+
+time_t
+mktime_no_dst(struct tm *t)
+/* same as mktime(), but without daylight saving time (dst) change adjustment
+ * (ie. the returned time_t does not depend on the tm_isdst field) */
+/* Remark : you may think than instead of creating a new function,
+ *          it would be easier to set the tm_isdst field to -1.
+ *          Unfortunately, the behaviour of mktime() with 
+ *          tm_isdst set to -1 depends on the unix you run.
+ *          In other word, it wouldn't be portable. */
+/*
+ * WARNING : the content of t has to be valide (for instance, 0<=t->tm_hour<=23,
+ *           etc)
+ */
+{
+    struct tm t2;
+    time_t ti1;
+
+    t2 = *t;
+
+    ti1 = mktime(&t2);
+    /* */
+    debug("after  mktime() : %d:%d isdst:%d ti:%ld\n",
+	  t2.tm_hour, t2.tm_min, t2.tm_isdst, ti1);
+    /* */
+
+    /* check if there have been a dst change adjustment */
+    if ( t->tm_isdst != t2.tm_isdst ) {
+
+	time_t ti2;
+	struct tm t3;
+
+	/* recompute the time_t field with the other isdst value
+	 * it works well, unless in a special case, hence the test
+	 * below */
+	t3 = *t;
+	t3.tm_isdst = t2.tm_isdst;
+	ti2 = mktime(&t3);
+	/* */
+	debug("after dst fix 1 : %d:%d isdst:%d ti:%ld\n",
+	       t3.tm_hour, t3.tm_min, t3.tm_isdst, ti2);
+	/* */
+
+	/* if t1 is in the "gap" of a dst change (for instance,
+	 * if t1 is 2:30 while at 2:00, it is 3:00 due to the dst change,
+	 * ie. 2:30 is never reached), the ti2 may be incorrect :
+	 * we check that it is correct before using it : */
+	if ( t3.tm_hour == t->tm_hour || ti1 < ti2 ) {
+	    t2 = t3;
+	    ti1 = ti2;
+	}
+
+    }
+
+    *t = t2;
+    return ti1;
+}
+
+
+void
+run_normal_job(cl_t *line, int info_fd)
+/* run a job, and write "log" on info_fd if positive */
 {
 
     if (line->cl_numexe <= 0 || 
 	(is_exe_sev(line->cl_option) && line->cl_numexe < UCHAR_MAX)) {
 	line->cl_numexe += 1;
 	run_queue_job(line);
+	send_msg_fd(info_fd, "Job %s started.", line->cl_shell);
     }
     else {
-	warn("    process already running: %s %s", 
-	     line->cl_file->cf_user,
-	     line->cl_shell
-	    );
+	warn_fd(info_fd, "    process already running: %s's %s", 
+		line->cl_file->cf_user,	line->cl_shell);
     } 
 
 }
@@ -113,8 +232,8 @@ run_serial_job(void)
 /*      debug("running next serial job"); */
 /*      //     */
 
-/*      debug("num: %d running:%d  index:%d", serial_num, serial_running, */
-/*  	  serial_array_index); */
+    debug("num: %d running:%d  index:%d", serial_num, serial_running,
+  	  serial_array_index);
     if ( serial_num != 0 ) {
 	run_queue_job(serial_array[serial_array_index]);
 	serial_array[serial_array_index] = NULL;
@@ -128,8 +247,27 @@ run_serial_job(void)
 }
 		    
 
+void 
+resize_exe_array(void)
+    /* make exe_array bigger */
+{
+    struct exe_t *ptr = NULL;
+    short int old_size = exe_array_size;
+
+    debug("Resizing exe_array");
+    exe_array_size = (exe_array_size + EXE_GROW_SIZE);
+	
+    if ( (ptr = calloc(exe_array_size, sizeof(struct exe_t))) == NULL )
+	die_e("could not calloc exe_array");
+
+    memcpy(ptr, exe_array, (sizeof(struct exe_t) * old_size));
+    free(exe_array);
+    exe_array = ptr;
+}
+
+
 void
-run_queue_job(CL *line)
+run_queue_job(cl_t *line)
     /* run a job */
 {
 
@@ -138,20 +276,9 @@ run_queue_job(CL *line)
 /*      // */
 
     /* append job to the list of executed job */
-    if ( exe_num >= exe_array_size ) {
-	struct exe *ptr = NULL;
-	short int old_size = exe_array_size;
+    if ( exe_num >= exe_array_size )
+	resize_exe_array();
 
-	debug("Resizing exe_array");
-	exe_array_size = (exe_array_size + EXE_GROW_SIZE);
-	
-	if ( (ptr = calloc(exe_array_size, sizeof(struct exe))) == NULL )
-	    die_e("could not calloc exe_array");
-
-	memcpy(ptr, exe_array, (sizeof(struct exe) * old_size));
-	free(exe_array);
-	exe_array = ptr;
-    }
     exe_array[exe_num].e_line = line;
 
     run_job(&exe_array[exe_num++]);
@@ -161,15 +288,15 @@ run_queue_job(CL *line)
 
 
 void
-insert_nextexe(CL *line)
+insert_nextexe(cl_t *line)
     /* insert a job the queue list */
 {
-    struct job *newjob;
+    struct job_t *newjob;
 
     if (queue_base != NULL) {
-	struct job *j;
-	struct job *jprev = NULL;
-	struct job *old_entry = NULL;
+	struct job_t *j;
+	struct job_t *jprev = NULL;
+	struct job_t *old_entry = NULL;
 
 	/* find the job in the list */
 	for (j = queue_base; j != NULL ; j = j->j_next)
@@ -199,7 +326,7 @@ insert_nextexe(CL *line)
 
 	if (old_entry == NULL) {
 	    /* this job wasn't in the queue : we append it */
-	    Alloc(newjob, job);
+	    Alloc(newjob, job_t);
 	    newjob->j_line = line;
 	}
 	else
@@ -216,7 +343,7 @@ insert_nextexe(CL *line)
     }
     else {
 	/* no job in queue */
-	Alloc(newjob, job);
+	Alloc(newjob, job_t);
 	newjob->j_line = line;	    
 	queue_base = newjob;
     }
@@ -224,39 +351,45 @@ insert_nextexe(CL *line)
 }
 
 void
-add_serial_job(CL *line)
+add_serial_job(cl_t *line, int info_fd)
     /* add the next queued job in serial queue */
 {
     short int i;
 
-    /* check if the line is already in the serial queue */
+    /* check if the line is already in the serial queue 
+     * (we consider serial jobs currently running as in the queue) */
     if ( (is_serial_sev(line->cl_option) && line->cl_numexe >= UCHAR_MAX) ||
-	 line->cl_numexe > 0) {
-	debug("already in serial queue %s", line->cl_shell);
+	 (! is_serial_sev(line->cl_option) && line->cl_numexe > 0) ) {
+	send_msg_fd_debug(info_fd, "already in serial queue %s", line->cl_shell);
 	return;
     }
 
-/*      debug("inserting in serial queue %s", line->cl_shell); */
+    send_msg_fd_debug(info_fd, "inserting in serial queue %s", line->cl_shell);
 
     if ( serial_num >= serial_array_size ) {
-	if ( serial_num >= SERIAL_QUEUE_MAX )
-	    /* run next job in the queue before adding the new one */
-	    run_serial_job();
+	if ( serial_num >= serial_queue_max ) {
+  	    error_fd(info_fd, "Could not add job : serial queue is full "
+		     "(%d jobs). Consider using option serialonce, fcron's "
+		     "option -m and/or -q : %s",serial_queue_max,line->cl_shell);
+	    if ( is_notice_notrun(line->cl_option) )
+		mail_notrun(line, QUEUE_FULL, NULL);
+	    return;
+	}
 	else {
-	    CL **ptr = NULL;
+	    cl_t **ptr = NULL;
 	    short int old_size = serial_array_size;
 
 	    debug("Resizing serial_array");
 	    serial_array_size = (serial_array_size + SERIAL_GROW_SIZE);
 	
-	    if ( (ptr = calloc(serial_array_size, sizeof(CL *))) == NULL )
+	    if ( (ptr = calloc(serial_array_size, sizeof(cl_t *))) == NULL )
 		die_e("could not calloc serial_array");
 
 	    /* copy lines in order to have the first line at the index 0 */
 	    memcpy(ptr + serial_array_index, serial_array,
-		   (sizeof(CL*) * (old_size - serial_array_index)) );
+		   (sizeof(cl_t*) * (old_size - serial_array_index)) );
 	    memcpy(ptr, serial_array + (old_size - serial_array_index),
-		   (sizeof(CL*) * serial_array_index));
+		   (sizeof(cl_t*) * serial_array_index));
 	    serial_array_index = 0;
 	    free(serial_array);
 	    serial_array = ptr;
@@ -271,54 +404,53 @@ add_serial_job(CL *line)
     serial_num++;
     line->cl_numexe += 1;
 
-    debug("serial num: %d size:%d index:%d curline:%d running:%d (%s)",
-	  serial_num, serial_array_size, serial_array_index, i,
-	  serial_running, line->cl_shell);
+    send_msg_fd_debug(info_fd, "serial num: %d size:%d index:%d curline:%d "
+		      "running:%d (%s)", serial_num, serial_array_size,
+		      serial_array_index, i, serial_running, line->cl_shell);
 
 
 }
 
 
 void
-add_lavg_job(CL *line)
+add_lavg_job(cl_t *line, int info_fd)
     /* add the next queued job in lavg queue */
     /* WARNING : must be run before a set_next_exe() to get the strict option
      * working correctly */
 {
 
-    /* check if the line is already in the lavg queue */
-    if ( (is_lavg_sev(line->cl_option) && line->cl_numexe >= UCHAR_MAX)
-	 || line->cl_numexe > 0 ) {
-	debug("already in lavg queue %s", line->cl_shell);
+    /* check if the line is already in the lavg queue
+     * (we consider serial jobs currently running as in the queue) */
+    if ( (is_lavg_sev(line->cl_option) && line->cl_numexe >= UCHAR_MAX) ||
+	 (! is_lavg_sev(line->cl_option) &&  line->cl_numexe > 0 ) ) {
+	send_msg_fd_debug(info_fd, "already in lavg queue %s", line->cl_shell);
 	return;
     }
 /*  	// */
-    debug("inserting in lavg queue %s", line->cl_shell);
+    send_msg_fd_debug(info_fd, "inserting in lavg queue %s", line->cl_shell);
 /*  	// */
 	
     /* append job to the list of lavg job */
     if ( lavg_num >= lavg_array_size ) {
-	if ( lavg_num >= LAVG_QUEUE_MAX ) {
-	    /* run the next lavg job (the oldest one) */
-	    register int i;
-	    int j = 0;
-		
-	    for (i = 1; i < lavg_num; i++)
-		if ( lavg_array[i].l_until < lavg_array[j].l_until )
-		    j = i;
-	    run_lavg_job(j);
+	if ( lavg_num >= lavg_queue_max ) {
+  	    error_fd(info_fd, "Could not add job : lavg queue is full (%d jobs)."
+		     " Consider using options lavgonce, until, strict and/or "
+		     "fcron's option -q.", lavg_queue_max, line->cl_shell);
+	    if ( is_notice_notrun(line->cl_option) )
+		mail_notrun(line, QUEUE_FULL, NULL);
+	    return;
 	}
 	else {
-	    struct lavg *ptr = NULL;
+	    struct lavg_t *ptr = NULL;
 	    short int old_size = lavg_array_size;
 		
 	    debug("Resizing lavg_array");
 	    lavg_array_size = (lavg_array_size + LAVG_GROW_SIZE);
 		
-	    if ( (ptr = calloc(lavg_array_size, sizeof(lavg))) == NULL )
+	    if ( (ptr = calloc(lavg_array_size, sizeof(lavg_t))) == NULL )
 		die_e("could not calloc lavg_array");
 		
-	    memcpy(ptr, lavg_array, (sizeof(lavg) * old_size));
+	    memcpy(ptr, lavg_array, (sizeof(lavg_t) * old_size));
 	    free(lavg_array);
 	    lavg_array = ptr;
 	}
@@ -330,18 +462,30 @@ add_lavg_job(CL *line)
     if ( is_strict(line->cl_option) && line->cl_runfreq == 1) {
 	struct tm *ft;
 	struct tm ftime;
-	time_t end_of_cur_int = 0;
+	time_t begin_of_cur_int, end_of_cur_int = 0;
+	int tz_changed = 0;
 
-	ft = localtime(&line->cl_nextexe);
+	/* Switch to another timezone if necessary. */
+	/* If line should be scheduled in a different time zone
+	 * (ie. cl_tz != NULL),
+	 * switch to that timezone now, do the calculations,
+	 * and switch back to the local timezone at the end 
+	 * of the function. */
+	tz_changed = switch_timezone(orig_tz_envvar, line->cl_tz);
+    
+	/* handle timezone differences */
+	begin_of_cur_int = line->cl_nextexe - (line->cl_file->cf_tzdiff*3600);
 
-	/* localtime() function seem to return every time the same pointer :
+	ft = localtime( &begin_of_cur_int );
+
+	/* localtime() function seems to return every time the same pointer :
 	   it resets our previous changes, so we need to prevent it
 	   ( localtime() is used in the debug() function) */
 	memcpy(&ftime, ft, sizeof(struct tm));
 
 	goto_non_matching(line, &ftime, END_OF_INTERVAL);
 
-	end_of_cur_int = mktime(&ftime);
+	end_of_cur_int = mktime_no_dst(&ftime) + (line->cl_file->cf_tzdiff * 3600);
 
 	if ((line->cl_until > 0) && (line->cl_until + now < end_of_cur_int))
 	    lavg_array[lavg_num].l_until = line->cl_until + now;
@@ -349,6 +493,9 @@ add_lavg_job(CL *line)
 	    lavg_array[lavg_num].l_until = end_of_cur_int;
 	    clear_run_if_late(line->cl_option);
 	}
+
+	if ( tz_changed > 0 )
+	    switch_back_timezone(orig_tz_envvar);
     }
     else
 	lavg_array[lavg_num].l_until = 
@@ -364,7 +511,7 @@ wait_chld(void)
 {
     short int i = 0;
     int pid;
-    CL *line = NULL;
+    cl_t *line = NULL;
 
 
 /*      // */
@@ -374,7 +521,7 @@ wait_chld(void)
     while ( (pid = wait3(NULL, WNOHANG, NULL)) > 0 ) {
 	i = 0;
 	while ( i < exe_num ) {
-	    if (pid == exe_array[i].e_pid) {
+	    if (pid == exe_array[i].e_ctrl_pid) {
 		if ( exe_array[i].e_line == NULL ) {
 		    /* the corresponding file has been removed from memory */
 		    debug("job finished: pid %d", pid);
@@ -388,12 +535,12 @@ wait_chld(void)
 		    
 		    if ( is_serial_once(line->cl_option) ) {
 			clear_serial_once(line->cl_option);
-			if ( --serial_running <= 0 )
+			if ( --serial_running < serial_max_running )
 			    run_serial_job();
 		    }
 		    else if ( is_serial(line->cl_option)
 			      && ! is_lavg(line->cl_option) ) {
-			if (--serial_running <= 0)
+			if (--serial_running < serial_max_running)
 			    run_serial_job();
 		    }
 		    else if ( is_lavg(line->cl_option) && 
@@ -407,13 +554,10 @@ wait_chld(void)
 		else
 		    exe_array[i].e_line = NULL;
 		
-		goto nextloop;
+		break;
 	    }
 	    i++;
 	}
-	/* execution shouldn't come here */
-	debug("not in exe_array !");
-      nextloop:
     }
 
 }
@@ -431,7 +575,7 @@ wait_all(int *counter)
     while ( (*counter > 0) && (pid = wait3(NULL, 0, NULL)) > 0 ) {
 	i = 0;
 	while ( i < exe_num ) {
-	    if (pid == exe_array[i].e_pid) {
+	    if (pid == exe_array[i].e_ctrl_pid) {
 		if ( exe_array[i].e_line == NULL ) {
 		    /* the corresponding file has been removed from memory */
 		    debug("job finished: pid %d", pid);
@@ -453,13 +597,10 @@ wait_all(int *counter)
 		else
 		    exe_array[i].e_line = NULL;
 		
-		goto nextloop;
+		break;
 	    }
 	    i++;
 	}
-	/* execution shouldn't come here */
-	debug("not in exe_array !");
-      nextloop:
     }    
     
 }
@@ -505,7 +646,7 @@ set_wday(struct tm *date)
    * and add this number modulo 7 to the wday number */
 {
     long nod = 0;
-    register int i;
+    int i;
 
     /* we add the number of days of each previous years */
     for (i =  (date->tm_year - 1); i >= 100 ; i--)
@@ -532,7 +673,7 @@ set_wday(struct tm *date)
 
 
 void
-goto_non_matching(CL *line, struct tm *ftime, char option)
+goto_non_matching(cl_t *line, struct tm *ftime, char option)
     /* search the first the nearest time and date that does
      * not match the line */
 {
@@ -570,12 +711,16 @@ goto_non_matching(CL *line, struct tm *ftime, char option)
 			} else
 			    ftime->tm_mday += to_add;
 		    } else {
-			ftime->tm_mday = 1;
 			if (is_freq_mons(line->cl_option)) {
 			    if (ftime->tm_mday >= 15)
 				ftime->tm_mon++;
 			    ftime->tm_mday = 15;
 			}
+			else
+			    /* weird : we have the bit freq_mid set, but
+			     * none of freq_{mins|hour|days|dow|mons} is set :
+			     * we does nothing */
+			    ;
 		    }
 		}
 	    }
@@ -632,10 +777,11 @@ goto_non_matching(CL *line, struct tm *ftime, char option)
 	if (option != END_OF_INTERVAL) {
 	    if (debug_opt)
 		set_wday(ftime);
-	    debug("   %s beginning of next period %d/%d/%d wday:%d %02d:%02d",
-		  line->cl_shell, (ftime->tm_mon + 1), ftime->tm_mday,
-		  (ftime->tm_year + 1900), ftime->tm_wday,
-		  ftime->tm_hour, ftime->tm_min);
+	    debug("   %s beginning of next period %d/%d/%d wday:%d %02d:%02d "
+		  "(tzdiff=%d, timezone=%s)", line->cl_shell, (ftime->tm_mon + 1),
+		  ftime->tm_mday, (ftime->tm_year + 1900), ftime->tm_wday,
+		  ftime->tm_hour, ftime->tm_min, line->cl_file->cf_tzdiff,
+		  (line->cl_tz != NULL)? line->cl_tz : "localtime");
 
 	    return;
 	}
@@ -756,44 +902,67 @@ goto_non_matching(CL *line, struct tm *ftime, char option)
 	    }
 	}
 	
-	debug("   %s %s %d/%d/%d wday:%d %02d:%02d", line->cl_shell,
+	debug("   %s %s %d/%d/%d wday:%d %02d:%02d (tzdiff=%d, timezone=%s)",
+	      line->cl_shell,
 	      (option == STD) ? "first non matching" : "end of interval",
-	      (ftime->tm_mon + 1), ftime->tm_mday,
-	      (ftime->tm_year + 1900), ftime->tm_wday,
-	      ftime->tm_hour, ftime->tm_min);
+	      (ftime->tm_mon + 1), ftime->tm_mday, (ftime->tm_year + 1900),
+	      ftime->tm_wday, ftime->tm_hour, ftime->tm_min,
+	      line->cl_file->cf_tzdiff,
+	      (line->cl_tz != NULL)? line->cl_tz : "localtime");
 	return;
     }
 }
 
 
 void 
-set_next_exe(CL *line, char option)
-  /* set the cl_nextexe of a given CL and insert it in the queue */
+set_next_exe(cl_t *line, char option, int info_fd)
+  /* set the cl_nextexe of a given cl_t and insert it in the queue */
 {
 
+    time_t basetime;
+    struct tm *ft;
+    struct tm ftime;
+    int tz_changed = 0;
+
+    basetime = (option & FROM_CUR_NEXTEXE) ? line->cl_nextexe : now ;
+
+    /* Switch to another timezone if necessary. */
+    /* If line should be scheduled in a different time zone
+     * (ie. cl_tz != NULL),
+     * switch to that timezone now, do the calculations,
+     * and switch back to the local timezone at the end 
+     * of the function. */
+    tz_changed = switch_timezone(orig_tz_envvar, line->cl_tz);
+
     if ( is_td(line->cl_option) ) {
 
-	struct tm *ft;
-	struct tm ftime;
-	register int i;
+	time_t nextexe = 0;
+	int i;
 	int max;
-	register char has_changed = 0;
+	char has_changed = 0;
 	/* to prevent from invinite loop with unvalid lines : */
 	short int year_limit = MAXYEAR_SCHEDULE_TIME;
+	/* timezone difference */
+	time_t basetime_tz = basetime - (line->cl_file->cf_tzdiff * 3600);
 
-	ft = localtime(&now);
+	ft = localtime(&basetime_tz);
 
 	/* localtime() function seem to return every time the same pointer :
 	   it resets our previous changes, so we need to prevent it
 	   ( localtime() is used in the debug() function) */
 	memcpy(&ftime, ft, sizeof(struct tm));
 
-	ftime.tm_isdst = -1;
+	/* creates a bug on DST change on some systems ?? */
+	/* ftime.tm_isdst = -1; */
 
-	/* to prevent multiple execution in the same minute */
-	ftime.tm_min += 1;
-	ftime.tm_sec = 0;
-	if (line->cl_runfreq==1 && option != NO_GOTO && option != NO_GOTO_LOG)
+	/* to prevent multiple execution of &-jobs in the same minute
+ 	 * (but not if the user has explicitely asked to run jobs immediately) */
+ 	if (first_sleep > 0 || option == STD || line->cl_runfreq != 1) {
+	    ftime.tm_min += 1;
+	    ftime.tm_sec = 0;
+	}
+	    
+	if (line->cl_runfreq == 1 && option != NO_GOTO && option != NO_GOTO_LOG)
 	    goto_non_matching(line, &ftime, STD);
 
       setMonth:
@@ -864,7 +1033,7 @@ set_next_exe(CL *line, char option)
 		goto setDay;
 	    }
 	} else {  /* dayor */
-	    register int j;
+	    int j;
 
 	    set_wday(&ftime);
 
@@ -924,51 +1093,127 @@ set_next_exe(CL *line, char option)
 	}
 	ftime.tm_min = i;
    
-      set_cl_nextexe:    
-	line->cl_nextexe = mktime(&ftime);
+      set_cl_nextexe:
+	/* set cl_nextexe (handle the timezone differences) */
+
+	/* NOTE : the output of mktime does not depend on the timezone,
+	 *        hence, nextexe is correct even if option timezone is used. */
+	nextexe = mktime_no_dst(&ftime);
+
+	if ( is_random(line->cl_option) ) {
+	    /* run the job at a random time during its interval of execution */
+	    struct tm intend;
+	    time_t intend_int;
+
+	    debug("   cmd: %s begin int exec %d/%d/%d wday:%d %02d:%02d "
+		  "(tzdiff=%d, timezone=%s)", line->cl_shell, (ftime.tm_mon + 1),
+		  ftime.tm_mday, (ftime.tm_year + 1900), ftime.tm_wday,
+		  ftime.tm_hour, ftime.tm_min, line->cl_file->cf_tzdiff,
+		  (line->cl_tz != NULL)? line->cl_tz : "localtime");
+
+	    memcpy(&intend, &ftime, sizeof(intend));
+	    goto_non_matching(line, &intend, END_OF_INTERVAL);
+	    intend_int = mktime_no_dst(&intend);
+
+	    /* set a random time to add to the first allowed time of execution */
+	    nextexe += ( (i= intend_int - nextexe) > 0) ? 
+		(time_t)(((float)i * (float)rand())/(float)RAND_MAX) : 0;
+	}
+
+	line->cl_nextexe = nextexe + (line->cl_file->cf_tzdiff * 3600);
+
+	if ( option != NO_GOTO ) {
+	    if ( is_random(line->cl_option) ) {
+		ft = localtime(&nextexe);
+		memcpy(&ftime, ft, sizeof(ftime));
+	    }
+	    send_msg_fd_debug(info_fd, "   cmd: %s next exec %d/%d/%d wday:%d "
+			      "%02d:%02d (tzdiff=%d, timezone=%s)", line->cl_shell,
+			      (ftime.tm_mon + 1), ftime.tm_mday,
+			      (ftime.tm_year + 1900), ftime.tm_wday,
+			      ftime.tm_hour, ftime.tm_min,
+			      line->cl_file->cf_tzdiff,
+			      (line->cl_tz != NULL)? line->cl_tz : "system's");
+	}
+
+	/* 
+	 * sanity check : 
+	 * if the nextexe is set to the past because of a bug,
+	 * the line will be executed again immediately, and it is most likely
+	 * to be set again in the past next time.
+	 * It would create a nasty infinite loop, a kind of "while(1) fork();"
+	 *
+	 * We add a test here to limit the consequences that would have
+	 * an unknown bug in this function.
+	 */
+	if ( line->cl_nextexe <= now ) {
+	    error("BUG ??? Fcron thinks the next exe time of %s is %ld, "
+		  "hence before now (%ld). To avoid infinite loop, nextexe"
+		  " will be set at now+5s.", line->cl_shell, line->cl_nextexe);
+	    line->cl_nextexe = now + 5;
+	}
 
-	if ( option != NO_GOTO )
-	    debug("   cmd: %s next exec %d/%d/%d wday:%d %02d:%02d",
-		  line->cl_shell, (ftime.tm_mon + 1), ftime.tm_mday,
-		  (ftime.tm_year + 1900), ftime.tm_wday,
-		  ftime.tm_hour, ftime.tm_min);
-	
     }
     else {
 	/* this is a job based on system up time */
-	line->cl_nextexe = now + line->cl_timefreq;
-	debug("   cmd: %s next exec in %lds", line->cl_shell,
-	      line->cl_timefreq);
+	line->cl_nextexe = basetime + line->cl_timefreq;
+
+	ft = localtime( &(line->cl_nextexe) );
+
+	/* localtime() function seem to return every time the same pointer :
+	   it resets our previous changes, so we need to prevent it
+	   ( localtime() is used in the debug() function) */
+	memcpy(&ftime, ft, sizeof(struct tm));
+
+	send_msg_fd_debug(info_fd, "   cmd: %s next exec %d/%d/%d wday:%d "
+			  "%02d:%02d (system time)", line->cl_shell,
+			  (ftime.tm_mon + 1), ftime.tm_mday, (ftime.tm_year+1900),
+			  ftime.tm_wday, ftime.tm_hour, ftime.tm_min);
     }
     
     insert_nextexe(line);
 
+    if ( tz_changed > 0 )
+	switch_back_timezone(orig_tz_envvar);
+
 }
 
 
 void
-set_next_exe_notrun(CL *line, char context)
+set_next_exe_notrun(cl_t *line, char context)
     /* set the time of the next execution and send a mail to tell user his job
      * has not run if necessary */
 {
-    time_t next_period = 0;
+    time_t previous_period = 0, next_period = 0;
     struct tm *ft = NULL;
     struct tm ftime, last_nextexe;
     char set_next_exe_opt = 0;
+    int tz_changed = 0;
 
 /*  // */
     debug("  set_next_exe_notrun : %s %d", line->cl_shell, context);
 /*  // */
 
+
+    /* Switch to another timezone if necessary. */
+    /* If line should be scheduled in a different time zone
+     * (ie. cl_tz != NULL),
+     * switch to that timezone now, do the calculations,
+     * and switch back to the local timezone at the end 
+     * of the function. */
+    tz_changed = switch_timezone(orig_tz_envvar, line->cl_tz);
+    
     if (context == SYSDOWN) {
-	ft = localtime(&line->cl_nextexe);
+	/* handle timezone differences */
+	previous_period = line->cl_nextexe - (line->cl_file->cf_tzdiff * 3600);
 	set_next_exe_opt = NO_GOTO;
     }
     else {
-	ft = localtime(&now);
+	previous_period = now - (line->cl_file->cf_tzdiff * 3600);
 	set_next_exe_opt = NO_GOTO_LOG;
     }
-    
+    ft = localtime(&previous_period);
+
     /* localtime() function seem to return every time the same pointer :
        it resets our previous changes, so we need to prevent it
        ( localtime() is used in the debug() function) */
@@ -978,24 +1223,27 @@ set_next_exe_notrun(CL *line, char context)
     
     ftime.tm_sec = 0;
     goto_non_matching(line, &ftime, STD);
-    next_period = mktime(&ftime);
+    next_period = mktime_no_dst(&ftime) + (line->cl_file->cf_tzdiff * 3600);
 
-    set_next_exe(line, set_next_exe_opt);
+    set_next_exe(line, set_next_exe_opt, -1);
     if ( line->cl_nextexe >= next_period ) {
 	/* line has not run during one or more period(s) : send a mail */
 	mail_notrun(line, context, &last_nextexe);
     }
 
+    if ( tz_changed > 0 )
+	switch_back_timezone(orig_tz_envvar);
+
 }
 
 void
-mail_notrun(CL *line, char context, struct tm *since)
-    /* send a mail to tell user a %-job has not run (and why) */
+mail_notrun(cl_t *line, char context, struct tm *since)
+    /* send a mail to tell user a job has not run (and why) */
 {
     int pid = 0;
-    int fd = 0;
-    struct tm *next2 = NULL, next;
-    char buf[LINE_LEN];
+    FILE *mailf = 0;
+    struct tm *time2 = NULL, time;
+    int tz_changed = 0;
 
     switch ( pid = fork() ) {
     case -1:
@@ -1012,70 +1260,79 @@ mail_notrun(CL *line, char context, struct tm *since)
 	      line->cl_shell, pid);
 /*  // */
 
-	/* create a entry in exe_array */
-	if ( exe_num >= exe_array_size ) {
-	    struct exe *ptr = NULL;
-	    short int old_size = exe_array_size;
-	    
-	    debug("Resizing exe_array");
-	    exe_array_size = (exe_array_size + EXE_GROW_SIZE);
-	    
-	    if ( (ptr = calloc(exe_array_size, sizeof(struct exe))) == NULL )
-		die_e("could not calloc exe_array");
-	    
-	    memcpy(ptr, exe_array, (sizeof(struct exe) * old_size));
-	    free(exe_array);
-	    exe_array = ptr;
-	}
+	/* create an entry in exe_array */
+	if ( exe_num >= exe_array_size )
+	    resize_exe_array();
 	/* set line to NULL as this is not a line ... */
 	exe_array[exe_num].e_line = NULL;
-	exe_array[exe_num].e_pid = pid;
+	exe_array[exe_num].e_ctrl_pid = pid;
 	exe_num++;
 	return;
     }
 
-    next2 = localtime(&line->cl_nextexe);
-    memcpy(&next, next2, sizeof(next));
+    /* Switch to another timezone if necessary. */
+    /* If line should be scheduled in a different time zone
+     * (ie. cl_tz != NULL),
+     * switch to that timezone now, do the calculations,
+     * and switch back to the local timezone at the end 
+     * of the function. */
+    tz_changed = switch_timezone(orig_tz_envvar, line->cl_tz);
+    
+    if ( context == QUEUE_FULL )
+	time2 = localtime(&now);
+    else
+	time2 = localtime(&line->cl_nextexe);
+    memcpy(&time, time2, sizeof(time));
 
     /* create a temp file, and write in it the message to send */
-    fd = create_mail(line, "Non-execution of fcron job");
+    mailf = create_mail(line, "Non-execution of fcron job");
 
-    if (context == SYSDOWN) {
-	snprintf(buf, sizeof(buf), "Line %s has not run since and including "
-		 "%d/%d/%d wday:%d %02d:%02d\ndue to system's down state.\n",
+    switch ( context ) {
+    case SYSDOWN:
+	fprintf(mailf, "Line %s has not run since and including "
+		"%d/%d/%d wday:%d %02d:%02d (timezone=%s)\n"
+		"due to system's down state.\n",
 		 line->cl_shell, (since->tm_mon + 1), since->tm_mday,
 		 (since->tm_year + 1900), since->tm_wday, since->tm_hour,
-		 since->tm_min);
-	xwrite(fd, buf);
-	snprintf(buf,sizeof(buf),"It will be next executed at %d/%d/%d wday:"
-		 "%d %02d:%02d\n", (next.tm_mon + 1), next.tm_mday,
-		 (next.tm_year+1900), next.tm_wday, next.tm_hour, next.tm_min);
-	xwrite(fd, buf);
-    }
-    else if (context == LAVG) {
-	snprintf(buf, sizeof(buf), "Line %s has not run since and including "
-		 "%d/%d/%d wday:%d %02d:%02d\n", line->cl_shell,
-		 (since->tm_mon + 1), since->tm_mday, (since->tm_year + 1900),
-		 since->tm_wday, since->tm_hour, since->tm_min);
-	xwrite(fd, buf);
-	snprintf(buf, sizeof(buf), "due to a too high system load average or "
+		 since->tm_min, (line->cl_tz)? line->cl_tz : "system's");
+	fprintf(mailf, "It will be next executed at %d/%d/%d wday:"
+		 "%d %02d:%02d\n", (time.tm_mon + 1), time.tm_mday,
+		 (time.tm_year+1900), time.tm_wday, time.tm_hour, time.tm_min);
+	break;
+    case LAVG:
+	fprintf(mailf, "Line %s has not run since and including "
+		 "%d/%d/%d wday:%d %02d:%02d (timezone=%s)\n",
+		line->cl_shell, (since->tm_mon + 1), since->tm_mday,
+		(since->tm_year + 1900), since->tm_wday, since->tm_hour,
+		since->tm_min, (line->cl_tz)? line->cl_tz : "system's");
+	fprintf(mailf, "due to a too high system load average or "
 		 "too many lavg-serial jobs.\n");
-	xwrite(fd, buf);	
-	snprintf(buf, sizeof(buf), "It will be next executed at %d/%d/%d "
-		 "wday:%d %02d:%02d\n", (next.tm_mon + 1), next.tm_mday,
-		 (next.tm_year+1900), next.tm_wday, next.tm_hour, next.tm_min);
-	xwrite(fd, buf);
+	fprintf(mailf, "It will be next executed at %d/%d/%d "
+		"wday:%d %02d:%02d (timezone=%s)\n", (time.tm_mon + 1),
+		time.tm_mday, (time.tm_year+1900), time.tm_wday, time.tm_hour,
+		time.tm_min, (line->cl_tz)? line->cl_tz : "system's");
+	break;
+    case QUEUE_FULL:
+	fprintf(mailf,"Line %s couldn't be added to lavg or serial queue which"
+		 " is full ( %d/%d/%d wday:%d %02d:%02d (timezone=%s)).\n",
+		line->cl_shell, (time.tm_mon + 1), time.tm_mday,
+		(time.tm_year + 1900), time.tm_wday, time.tm_hour, time.tm_min,
+		(line->cl_tz)? line->cl_tz : "system's");
+	fprintf(mailf, "Consider using options lavgonce, until, strict, "
+		 "serialonce and/or fcron's option -m.\n");
+	fprintf(mailf, "Note that job %s has not run.\n", line->cl_shell);
+	break;
     }
     
     /* become user (for security reasons) */
-    if (change_user(line->cl_runas) < 0)
+    if (change_user(line) < 0)
 	return ;
 
     /* then, send mail */
-    launch_mailer(line, fd);
+    launch_mailer(line, mailf);
     
     /* we should not come here : launch_mailer does not return */
-    error("mail_notrun : launch_mailer failed");
+    die("mail_notrun : launch_mailer failed");
     
 }
 
@@ -1089,15 +1346,13 @@ check_lavg(time_t lim)
 #ifdef NOLOADAVG
     while ( lavg_num > 0 )
 	run_lavg_job(0);
+
+    tts = time_to_sleep(lim);
     return tts;
 #else
 
-    register int i = 0;
-    double l_avg[3];
-
-    l_avg[0] = 0;
-    l_avg[1] = 0;
-    l_avg[2] = 0;
+    int i = 0;
+    double l_avg[3]= {0, 0, 0};
 
     /* first, check if some lines must be executed because of until */
     while ( i < lavg_num )
@@ -1110,10 +1365,11 @@ check_lavg(time_t lim)
 			    lavg_array[i].l_line->cl_shell);
 
 		/* set time of the next execution and send a mail if needed */
-		if ( is_notice_notrun(lavg_array[i].l_line->cl_option) )
+		if ( is_td(lavg_array[i].l_line->cl_option) &&
+		     is_notice_notrun(lavg_array[i].l_line->cl_option) )
   		    set_next_exe_notrun(lavg_array[i].l_line, LAVG);
 		else
-		    set_next_exe(lavg_array[i].l_line, NO_GOTO_LOG);
+		    set_next_exe(lavg_array[i].l_line, NO_GOTO_LOG, -1);
 
 		/* remove this job from the lavg queue */
 		lavg_array[i].l_line->cl_numexe -= 1;
@@ -1149,7 +1405,7 @@ check_lavg(time_t lim)
     i = 0;
     while ( i < lavg_num ) {
 	/* check if the line should be executed */
-	if ( lavg_serial_running > 0 && 
+	if ( lavg_serial_running >= serial_max_running && 
 	     is_serial(lavg_array[i].l_line->cl_option) ) {
 	    i++;
 	    continue;
@@ -1200,15 +1456,17 @@ time_to_sleep(time_t lim)
     time_t tts = lim;
     time_t ti = time(NULL);
 
+    /* note : jobs in queue_base are sorted */
     if ( queue_base != NULL ) {
 	if ( queue_base->j_line->cl_nextexe < lim )
 	    tts = queue_base->j_line->cl_nextexe;
     }
 	    
-    if ( (tts = tts - ti) < 0)
+    tts = tts - ti;
+    if ( tts < 0)
 	tts = 0;
 
-/*      // debug("Time to sleep: %lds", tts); */
+/*      debug("Time to sleep: %lds", tts); */
 
     return tts;
 }
