@@ -1,7 +1,7 @@
 /*
  * FCRON - periodic command scheduler 
  *
- *  Copyright 2000-2006 Thibault Godouet <fcron@free.fr>
+ *  Copyright 2000-2007 Thibault Godouet <fcron@free.fr>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
  *  `LICENSE' that comes with the fcron source distribution.
  */
 
- /* $Id: fileconf.c,v 1.76 2006/01/11 00:53:05 thib Exp thib $ */
+ /* $Id: fileconf.c,v 1.80 2007/04/14 18:04:15 thib Exp thib $ */
 
 #include "fcrontab.h"
 
@@ -62,6 +62,8 @@ const char *mons_ary[] = {
 };
 
 
+#define GET_LINE_EOF 999
+
 char *
 get_string(char *ptr)
     /* read string pointed by ptr, remove blanks and manage
@@ -126,7 +128,7 @@ get_line(char *str, size_t size, FILE *file)
 	    *(str + i) = (char) '\0';
 	    /* we couldn't return EOF ( equal to ERR by default ) 
 	     * nor ERR, which is used for another error */
-	    return 999;
+	    return GET_LINE_EOF;
 
 	default:
 	    *(str + i) = (char) c;
@@ -137,7 +139,7 @@ get_line(char *str, size_t size, FILE *file)
     }
 
     /* line is too long : goto next line and return ERR */
-    while ( ( (char)(c = getc(file)) != EOF ) && ( (char)c != '\n') )
+    while ( ( (c = getc(file)) != EOF ) && ( c != '\n') )
 	;
     line++;
     need_correction = 1;
@@ -186,7 +188,7 @@ read_file(char *filename)
     if ( debug_opt )
 	fprintf(stderr, "FILE %s\n", file_name);
 
-    if (strcmp(runas, "root") == 0)
+    if (strcmp(runas, ROOTNAME) == 0)
 	max_entries = 65535;
 
     /* max_lines acts here as a security counter to avoid endless loop. */
@@ -194,9 +196,11 @@ read_file(char *filename)
 
     while ( entries <= max_entries && line <= max_lines ) {
 
-	if ( (ret = get_line(buf, sizeof(buf), file)) == ERR ) {
+	ret = get_line(buf, sizeof(buf), file);
+
+	if ( ret == ERR ) {
 	    fprintf(stderr, "%s:%d: Line is too long (more than %d): skipping line.\n",
-		    file_name, line, sizeof(buf));
+		    file_name, line, (int)sizeof(buf));
 	    continue;
 	}
 
@@ -210,8 +214,7 @@ read_file(char *filename)
 	case '#':
 	case '\0':
 	    /* comments or empty line: skipping */
-	    line++;
-	    continue;
+	    break;
 	case '@':
 	    read_freq(ptr, cf);
 	    entries++;
@@ -243,7 +246,8 @@ read_file(char *filename)
 	line++;	
 
 	if ( ret != OK )
-	    /* in this case, ret == EOF : no more lines */
+	    /* in this case, ret == GET_LINE_EOF : 
+	     * no more lines, so we exit the loop */
 	    break;
 	
     }
@@ -376,7 +380,7 @@ get_nice(char *ptr, int *nice)
 	return NULL;
 
     if ( negative == 1 ) {
-	if (getuid() != ROOTUID) {
+	if (getuid() != rootuid) {
 	    fprintf(stderr, "must be privileged to use a negative argument "
 		    "with nice: set to 0\n");
 	    need_correction = 1;
@@ -842,7 +846,7 @@ read_opt(char *ptr, cl_t *cl)
 	}
 
 	else if(strcmp(opt_name, "runas") == 0) {
-	    if (getuid() != ROOTUID) {
+	    if (getuid() != rootuid) {
 		fprintf(stderr, "must be privileged to use option runas: "
 			"skipping option\n");
 		need_correction = 1;
@@ -1085,7 +1089,7 @@ check_username(char *ptr, cf_t *cf, cl_t *cl)
         ptr = ptr + indx;	/* move ptr to the next word */
 	Skip_blanks(ptr);
 
-	if (getuid() != ROOTUID) {
+	if (getuid() != rootuid) {
 	    fprintf(stderr, "must be privileged to run as another user : "
 		    "ignoring\n");
 	} else {
