@@ -1,7 +1,7 @@
 /*
  * FCRON - periodic command scheduler 
  *
- *  Copyright 2000-2001 Thibault Godouet <fcron@free.fr>
+ *  Copyright 2000-2007 Thibault Godouet <fcron@free.fr>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
  *  `LICENSE' that comes with the fcron source distribution.
  */
 
- /* $Id: option.h,v 1.16 2001/02/01 20:53:11 thib Exp $ */
+ /* $Id: option.h,v 1.25 2007/04/14 18:04:26 thib Exp thib $ */
 
 /* This has been inspired from bitstring(3) : here is the original copyright :
  */
@@ -47,8 +47,14 @@
  *
  */
 
+
 /* read and set options of a line */
 
+/* WARNING : do not change any option number, nor remove any option, nor change
+ *           the meaning of a value (i.e. bit set to 1 -> true).
+ *           It can lead to errors with fcrontabs save to disk/load in memory.
+ */
+
 /*
   
   The options are :
@@ -61,7 +67,7 @@
   4      should we run this job at fcron startup if it should have been
            executed during system down?
   5      should this job be run serially ?
-  6      is this job should be run serially only once (for bootrun) ?
+  6      should this job be run serially only once (for bootrun) ?
   7      does the output have to be mailed to user ?
   8      does the output (even if zero-length) must be mailed to user ?
   9      if time of execution is exceeded, exec the lavg job or not ?
@@ -78,11 +84,12 @@
   20     Should we remove a %-job from lavg queue if the interval is exceeded ?
   21     Should user be mailed if a %-job has not run during a period ?
   22     Should fcron log everything about this job or just errors ?
+  23     Should this job be run asap, or randomly in its allowed interval of execution ?
+  24     Should first value be applied at each fcron startup, or before line first exe ?
+  25     if fcron is running in foreground, print jobs output to stderr/stdout or mail ?
 
 */
 
-/* default value corresponds to a bit value of 0 */
-
 #ifndef __OPTIONH__
 #define __OPTIONH__
 
@@ -106,6 +113,16 @@
 
 /* external macros */
 
+/* default value generally corresponds to a bit value of 0 : if you want to
+ * change the default value of an option, do it by modifying the following
+ * macro (set to 1 the needed bits) */
+#define set_default_opt(opt)  \
+       { \
+          if ( SERIAL_ONCE >= 1 ) clear_serial_sev(opt); \
+          if ( LAVG_ONCE == 0 ) set_lavg_sev(opt); \
+       }
+
+
 /*
   bit 0 : set to 1 : line is based on system up time
           set to 0 : line is based on time and date
@@ -244,18 +261,6 @@
 	(_bit_clear(opt, 10))
 
 
-#if ( SERIAL_ONCE == 1 )
-/*
-  bit 11 : set to 1 : can be put several times in serial queue simultaneously
-           set to 0 : can only be put once in serial queue simultaneously
-*/
-#define	is_serial_sev(opt) \
-	(_bit_test(opt, 11))
-#define	set_serial_sev(opt) \
-	(_bit_set(opt, 11))
-#define clear_serial_sev(opt) \
-	(_bit_clear(opt, 11))
-#else
 /*
   bit 11 : set to 1 : can only be put once in serial queue simultaneously
            set to 0 : can be put several times in serial queue simultaneously
@@ -266,13 +271,11 @@
 	(_bit_clear(opt, 11))
 #define clear_serial_sev(opt) \
 	(_bit_set(opt, 11))
-#endif /* SERIAL_ONCE == 0 */
 
 
-#if ( LAVG_ONCE == 1 )
 /*
-  bit 12 : set to 1 : can be put several times in lavg queue simultaneously
-           set to 0 : can only be put once in lavg queue simultaneously
+  bit 12 : set to 1 : can only be put once in lavg queue simultaneously
+           set to 0 : can be put several times in lavg queue simultaneously
 */
 #define	is_lavg_sev(opt) \
 	(_bit_test(opt, 12))
@@ -280,18 +283,6 @@
 	(_bit_set(opt, 12))
 #define clear_lavg_sev(opt) \
 	(_bit_clear(opt, 12))
-#else
-/*
-  bit 12 : set to 1 : can only be put once in lavg queue simultaneously
-           set to 0 : can be put several times in lavg queue simultaneously
-*/
-#define	is_lavg_sev(opt) \
-	( ! _bit_test(opt, 12))
-#define	set_lavg_sev(opt) \
-	(_bit_clear(opt, 12))
-#define clear_lavg_sev(opt) \
-	(_bit_set(opt, 12))
-#endif /* LAVG_ONCE == 0 */
 
 
 /*
@@ -379,8 +370,8 @@
 
 
 /*
-  bit 20 : set to 1 : remove %-job from lavg queue if interval is exceeded
-           set to 0 : let the job in the %-queue if interval is exceeded
+  bit 20 : set to 1 : let the job in the %-queue if interval is exceeded
+           set to 0 : remove %-job from lavg queue if interval is exceeded
 */
 #define	is_strict(opt) \
 	( ! _bit_test(opt, 20))
@@ -414,5 +405,43 @@
 	(_bit_clear(opt, 22))
 
 
+/*
+  bit 23 : set to 1 : run this job at a random time in its allowed interval of execution.
+           set to 0 : run this job asap (safer)
+*/
+#define	is_random(opt) \
+	(_bit_test(opt, 23))
+#define	set_random(opt) \
+	(_bit_set(opt, 23))
+#define clear_random(opt) \
+	(_bit_clear(opt, 23))
+
+
+/*
+  bit 24 : set to 1 : "volatile" system up time, i.e. restart counting each time fcron
+                      is started
+           set to 0 : continue counting uptime where last fcron instance left of
+*/
+#define	is_volatile(opt) \
+	(_bit_test(opt, 24))
+#define	set_volatile(opt) \
+	(_bit_set(opt, 24))
+#define clear_volatile(opt) \
+	(_bit_clear(opt, 24))
+
+/*
+  bit 25 : set to 1 : if fcron is running in the forground, then also let jobs print
+                      to stderr/stdout instead of mailing or discarding it
+           set to 0 : if fcron is not running in the foreground or this bit is not
+	              set, then treat it as specified with the other options
+*/
+#define	is_stdout(opt) \
+	(_bit_test(opt, 25))
+#define	set_stdout(opt) \
+	(_bit_set(opt, 25))
+#define clear_stdout(opt) \
+	(_bit_clear(opt, 25))
+
+
 #endif /* __OPTIONH__ */
 
